/*
   FNT - Stream handler for socket connections
   Open socket connections (server or client)

   Copyright (C) 2009 Free Software Foundation, Inc.

   Written by:
   Slava Zanko <slavazanko@gmail.com>


   This file is part of the FIDO New Technology.

   FIDO New Technology is free software: you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as published 
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "../include/global.h"
#include "../include/types.h"
#include "../include/error.h"
#include "../include/stream.h"

#include "main.h"
#include "utils.h"

/*** global variables **************************************************/

/*** file scope macro definitions **************************************/

/*** file scope type declarations **************************************/

/*** file scope functions **********************************************/

static int
fnt_stream_open_socket_server (__fnt_stream_socketInfo_t * socketInfo)
{
    struct sockaddr_storage sa;
    int fd, flag = 1;

    memset (&sa, 0, sizeof (sa));

    if (fnt_stream_socketPton (socketInfo->host, &sa) == false)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_
			("Unable convert adress '%s' to internal network address representation"),
			socketInfo->host);
	return -1;
    }
    sa.ss_family = AF_INET;

    if ((fd = socket (sa.ss_family, SOCK_STREAM, 0)) < 0)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Unable to create socket on '%s'"),
			socketInfo->host);
	return -1;
    }

    if (setsockopt
	(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof (flag)) < 0)
	return -1;

    fnt_stream_socketSetPort (&sa, socketInfo->port);

    if (bind (fd, (struct sockaddr *) &sa, fnt_stream_socketAddrLen (&sa)) <
	0)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Unable to bind to '%s:%d'"), socketInfo->host,
			socketInfo->port);
	fnt_stream_close_socket (fd);
	return -1;
    }
    if (listen (fd, 10) < 0)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Unable to listen on '%s:%d'"), socketInfo->host,
			socketInfo->port);
	fnt_stream_close_socket (fd);
	return -1;
    }
    return fd;
}

static int
fnt_stream_open_socket_client (__fnt_stream_socketInfo_t * socketInfo)
{
    int fd, flag, flags;
    struct sockaddr_storage sa, src;

    memset (&sa, 0, sizeof (sa));

    if (fnt_stream_socketPton (socketInfo->host, &sa) == false)
	return -1;

    if ((fd = socket (sa.ss_family, SOCK_STREAM, 0)) < 0)
	return -1;

    fnt_stream_socketSetPort (&sa, socketInfo->port);
    flag =
	connect (fd, (struct sockaddr *) &sa,
		 fnt_stream_socketAddrLen (&sa));

    if (flag == 0)
	return fd;
    /*
       #ifdef G_OS_WIN32
       if (flag == -1 && WSAGetLastError () == WSAEWOULDBLOCK)
       #else
       if (flag == -1 && errno == EINPROGRESS)
       #endif
       {
       g_io_channel_set_flags(gio_stream,G_IO_FLAG_NONBLOCK,NULL);
       }
     */
    fnt_stream_close_socket (fd);
    return -1;
}

/*** public functions **************************************************/

bool
open_stream (fnt_stream_t * stream, const char *data)
{
    int fd;
    __fnt_stream_socketInfo_t socketInfo;
    __fnt_driver_data_t *driver_data;


    if (stream == NULL || data == NULL)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Wrong parameters"));
	return false;
    }

    if (stream->driver_data != NULL)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Stream '%s' already opened"), stream->init_data);
	return false;
    }


    if (!fnt_stream_parceSocketInfo (&socketInfo, data))
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_OPEN,
			_("Wrong connection info string: %s"), data);

	return false;
    }

    driver_data = g_try_malloc0 (sizeof (__fnt_driver_data_t));
    if (driver_data == NULL)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_INTERNAL,
			_("Unable to allocate memory for '%s'"),
			"__fnt_driver_data_t");
	g_free (socketInfo.host);
	return false;
    }

    stream->is_server = socketInfo.is_server;

    if (socketInfo.is_server)
    {
	driver_data->remote_host = g_strdup ("");
	driver_data->remote_port = -1;
	fd = fnt_stream_open_socket_server (&socketInfo);
    }
    else
    {
	driver_data->remote_host = g_strdup (socketInfo.host);
	driver_data->remote_port = socketInfo.port;
	fd = fnt_stream_open_socket_client (&socketInfo);
    }

    if (fd == -1)
    {
	fnt_error_push (FNT_MODULE_stream, FNT_STREAM_E_INTERNAL,
			_("Unable to open connection"));
	g_free(driver_data->remote_host);
	g_free (driver_data);
	g_free (socketInfo.host);
	return false;
    }

#ifdef G_OS_WIN32
    driver_data->socket = g_io_channel_win32_new_socket (fd);
#else
    driver_data->socket = g_io_channel_unix_new (fd);
#endif
    g_io_channel_set_flags (driver_data->socket, G_IO_FLAG_NONBLOCK, NULL);
    g_io_channel_set_encoding (driver_data->socket, NULL, NULL);
    g_io_channel_set_buffered (driver_data->socket, FALSE);
    g_free (socketInfo.host);

    stream->driver_data = (void *) driver_data;
    stream->state = FNT_STREAM_STATE_ok;

    return true;
}
